import logging
from typing import Dict

from peek_abstract_chunked_index.private.client.handlers.ACICacheHandlerABC import (
    ACICacheHandlerABC,
)
from peek_abstract_chunked_index.private.tuples.ACIUpdateDateTupleABC import (
    ACIUpdateDateTupleABC,
)
from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.client.controller.LocationIndexCacheController import (
    clientLocationIndexUpdateFromServerFilt,
)
from peek_plugin_diagram._private.tuples.location_index.LocationIndexUpdateDateTuple import (
    LocationIndexUpdateDateTuple,
)

logger = logging.getLogger(__name__)

clientLocationIndexWatchUpdateFromDeviceFilt = {
    "key": "clientLocationIndexWatchUpdateFromDevice"
}
clientLocationIndexWatchUpdateFromDeviceFilt.update(diagramFilt)


# ModelSet HANDLER
class LocationIndexCacheHandler(ACICacheHandlerABC):
    _UpdateDateTuple: ACIUpdateDateTupleABC = LocationIndexUpdateDateTuple
    _updateFromDeviceFilt: Dict = clientLocationIndexWatchUpdateFromDeviceFilt
    _updateFromLogicFilt: Dict = clientLocationIndexUpdateFromServerFilt
    _logger: logging.Logger = logger



import logging

from sqlalchemy import BigInteger
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import Index
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_abstract_chunked_index.private.tuples.ACIEncodedChunkTupleABC import (
    ACIEncodedChunkTupleABC,
)
from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import (
    ACIProcessorQueueTupleABC,
)
from peek_plugin_base.storage.TypeDecorators import PeekLargeBinary
from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from .DeclarativeBase import DeclarativeBase
from .Display import DispBase
from .ModelSet import ModelSet

logger = logging.getLogger(__name__)


@addTupleType
class LocationIndexCompilerQueue(
    DeclarativeBase, Tuple, ACIProcessorQueueTupleABC
):
    __tablename__ = "LocationIndexCompilerQueue"
    __tupleType__ = diagramTuplePrefix + __tablename__

    id = Column(BigInteger, primary_key=True, autoincrement=True)

    indexBucket = Column(String(100), primary_key=True)
    modelSetId = Column(
        Integer, ForeignKey("ModelSet.id", ondelete="CASCADE"), primary_key=True
    )

    __table_args__ = (
        Index(
            "idx_LICompQueue_modelSetId_indexBucket",
            modelSetId,
            indexBucket,
            unique=False,
        ),
    )

    # noinspection PyUnresolvedReferences
    @classmethod
    def sqlCoreLoad(cls, row):
        return LocationIndexCompilerQueue(
            id=row.id, modelSetId=row.modelSetId, indexBucket=row.indexBucket
        )

    @property
    def ckiUniqueKey(self):
        return self.indexBucket


@addTupleType
class LocationIndex(DeclarativeBase, Tuple):
    __tablename__ = "LocationIndex"
    __tupleType__ = diagramTuplePrefix + __tablename__

    indexBucket = Column(String(100), primary_key=True)
    dispId = Column(
        BigInteger,
        ForeignKey("DispBase.id", ondelete="CASCADE"),
        primary_key=True,
    )

    disp = relationship(DispBase)

    modelSetId = Column(Integer, ForeignKey("ModelSet.id"), nullable=False)
    modelSet = relationship(ModelSet)

    __table_args__ = (
        Index("idx_LocationIndex_indexBucket", indexBucket, unique=False),
        Index("idx_LocationIndex_dispId", dispId, unique=False),
        Index("idx_LocationIndex_modelSetId", modelSetId, unique=False),
    )


@addTupleType
class LocationIndexCompiled(DeclarativeBase, Tuple, ACIEncodedChunkTupleABC):
    __tablename__ = "LocationIndexCompiled"
    __tupleType__ = diagramTuplePrefix + __tablename__

    id = Column(BigInteger, primary_key=True, autoincrement=True)

    indexBucket = Column(String(100), primary_key=True)
    blobData = Column(PeekLargeBinary, nullable=False)
    lastUpdate = Column(String(50), nullable=False)

    modelSetId = Column(
        Integer, ForeignKey("ModelSet.id", ondelete="CASCADE"), nullable=False
    )
    modelSet = relationship(ModelSet)

    __table_args__ = (
        Index("idx_LIIndexUpdate_modelSetId", modelSetId, unique=False),
        Index("idx_LIIndexUpdate_indexBucket", indexBucket, unique=True),
    )

    @property
    def ckiChunkKey(self):
        return self.indexBucket

    @property
    def ckiEncodedData(self):
        return self.blobData

    @property
    def ckiHasEncodedData(self) -> bool:
        return bool(self.blobData)

    @property
    def ckiLastUpdate(self):
        return self.lastUpdate

    @classmethod
    def ckiCreateDeleteEncodedChunk(cls, chunkKey: str):
        from peek_plugin_diagram._private.tuples.location_index.EncodedLocationIndexTuple import (
            EncodedLocationIndexTuple,
        )

        return EncodedLocationIndexTuple(indexBucket=chunkKey)

    @classmethod
    def sqlCoreChunkKeyColumn(cls):
        return cls.__table__.c.indexBucket

    @classmethod
    def sqlCoreLastUpdateColumn(cls):
        return cls.__table__.c.lastUpdate

    @classmethod
    def sqlCoreLoad(cls, row):
        from peek_plugin_diagram._private.tuples.location_index.EncodedLocationIndexTuple import (
            EncodedLocationIndexTuple,
        )

        return EncodedLocationIndexTuple(
            modelSetKey=row.key,
            indexBucket=row.indexBucket,
            encodedLocationIndexTuple=row.blobData,
            lastUpdate=row.lastUpdate,
        )



"""set ModelCoordSet.order default

Peek Plugin Database Migration Script

Revision ID: 1e2f86ad5915
Revises: 18bd59a236c8
Create Date: 2024-10-22 18:56:42.900094

"""

# revision identifiers, used by Alembic.
revision = "1e2f86ad5915"
down_revision = "18bd59a236c8"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.alter_column(
        "ModelCoordSet",
        sa.Column("order", sa.Integer(), server_default="0", nullable=True),
        schema="pl_diagram",
    )


def downgrade():
    pass



"""added wrapTextAtCharSplitBetweenWords

Peek Plugin Database Migration Script

Revision ID: dcc9892e18a8
Revises: a151fd42572b
Create Date: 2023-08-04 09:51:29.490205

"""

# revision identifiers, used by Alembic.
revision = "dcc9892e18a8"
down_revision = "a151fd42572b"
branch_labels = None
depends_on = None

from alembic import (
    op,
)
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.add_column(
        "DispTextStyle",
        sa.Column(
            "wrapTextAtCharSplitBetweenWords",
            sa.Boolean(),
            server_default="true",
            nullable=True,
        ),
        schema="pl_diagram",
    )

    op.execute(
        """
        UPDATE pl_diagram."DispTextStyle"
        SET "wrapTextAtCharSplitBetweenWords" = true
        WHERE "wrapTextAtCharSplitBetweenWords" is null;
    """
    )

    op.alter_column(
        "DispTextStyle",
        "wrapTextAtCharSplitBetweenWords",
        nullable=False,
        schema="pl_diagram",
    )


def downgrade():
    op.drop_column(
        "DispTextStyle",
        "wrapTextAtCharSplitBetweenWords",
        schema="pl_diagram",
    )



"""make light mode optional

Peek Plugin Database Migration Script

Revision ID: c579df20f0f8
Revises: b88c518e8543
Create Date: 2023-10-12 14:07:37.315132

"""

# revision identifiers, used by Alembic.
revision = "c579df20f0f8"
down_revision = "b88c518e8543"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column(
        "ModelCoordSet",
        sa.Column(
            "lightModeEnabled",
            sa.Boolean(),
            nullable=True,
            server_default="true",
        ),
        schema="pl_diagram",
    )

    # set 'blockApiUpdate' to false
    op.execute(
        f"""
                UPDATE pl_diagram."ModelCoordSet"
                SET "lightModeEnabled" = true
                WHERE "lightModeEnabled" IS NULL;
            """
    )
    # set 'blockApiUpdate' NOT NULL
    op.alter_column(
        "ModelCoordSet", "lightModeEnabled", nullable=False, schema="pl_diagram"
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("ModelCoordSet", "lightModeEnabled", schema="pl_diagram")
    # ### end Alembic commands ###



import json
import logging
from collections import defaultdict
from typing import List
from typing import Union

from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_plugin_diagram._private.client.controller.BranchIndexCacheController import (
    BranchIndexCacheController,
)
from peek_plugin_diagram._private.storage.branch.BranchIndexEncodedChunk import (
    BranchIndexEncodedChunk,
)
from peek_plugin_diagram._private.tuples.branch.BranchTuple import BranchTuple
from peek_plugin_diagram._private.worker.tasks.branch._BranchIndexCalcChunkKey import (
    makeChunkKeyForBranchIndex,
)

logger = logging.getLogger(__name__)


class BranchTupleProvider(TuplesProviderABC):
    def __init__(self, cacheHandler: BranchIndexCacheController):
        self._cacheHandler = cacheHandler

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        modelSetKey = tupleSelector.selector["modelSetKey"]
        coordSetId = tupleSelector.selector["coordSetId"]
        keys = tupleSelector.selector["keys"]

        keysByChunkKey = defaultdict(list)

        results: List[BranchTuple] = []

        for key in keys:
            keysByChunkKey[makeChunkKeyForBranchIndex(modelSetKey, key)].append(
                key
            )

        for chunkKey, branchKeys in keysByChunkKey.items():
            chunk: BranchIndexEncodedChunk = self._cacheHandler.encodedChunk(
                chunkKey
            )

            if not chunk:
                logger.warning(
                    "BranchIndex chunk %s is missing from cache", chunkKey
                )
                continue

            resultsByKeyStr = (
                Payload().fromEncodedPayload(chunk.encodedData).tuples[0]
            )
            resultsByKey = json.loads(resultsByKeyStr)

            for branchKey in branchKeys:
                if branchKey not in resultsByKey:
                    logger.warning(
                        "Branch %s is missing from index, chunkKey %s",
                        branchKey,
                        chunkKey,
                    )
                    continue

                packedJsons = resultsByKey[branchKey]

                for packedJson in packedJsons:
                    result = BranchTuple.loadFromJson(packedJson, None, None)
                    if result.coordSetId == coordSetId or coordSetId is None:
                        results.append(result)

        # Create the vortex message
        return Payload(filt, tuples=results).makePayloadEnvelope().toVortexMsg()
